from itertools import chain

from .. import query
from ..graph import Link, Field, Nothing
from ..types import TypeRef, Sequence
from ..query import merge
from ..types import Unknown
from ..engine import Query, store_fields, subquery
from ..expr.refs import RequirementsExtractor
from ..expr.core import to_expr
from ..expr.checker import check, graph_types, fn_types
from ..expr.compiler import ExpressionCompiler


THIS_LINK_NAME = '__link_to_this'


def _create_result_proc(query, env, node, fields, procs, options, ids):
    def result_proc(result):
        sq_result = query.result()
        store_fields(result, node, fields, ids, [
            [proc(this, env, sq_result, *opt_args)
             for proc, opt_args in zip(procs, options)]
            for this in sq_result[THIS_LINK_NAME]
        ])
    return result_proc


class Expr(Field):

    def __init__(self, name, subquery, *other, **kwargs):
        if not len(other):
            raise TypeError('Missing required argument')
        elif len(other) == 1:
            type_, expr = None, other[0]
        elif len(other) == 2:
            type_, expr = other
        else:
            raise TypeError('More positional arguments ({}) than expected (2)'
                            .format(len(other)))

        super(Expr, self).__init__(name, type_, subquery, **kwargs)

        expr_node, functions = to_expr(expr)

        self.functions = functions

        types = subquery.types.copy()
        types.update(fn_types(self.functions))
        types.update({opt.name: Unknown for opt in self.options})

        expr_node = check(expr_node, types)

        options_set = set(self.options_map)
        query_node = RequirementsExtractor.extract(types, expr_node)
        query_node = query.Node([f for f in query_node.fields
                                 if f.name not in options_set])
        self.reqs = query_node

        option_names = [opt.name for opt in self.options]
        code = ExpressionCompiler.compile_lambda_expr(expr_node, option_names)
        self.proc = eval(compile(code, '<expr>', 'eval'))


def _yield_options(query_field, graph_field):
    options = query_field.options or {}
    for option in graph_field.options:
        value = options.get(option.name, option.default)
        if value is Nothing:
            raise TypeError('Required option "{}" for {!r} was not provided'
                            .format(option.name, graph_field))
        else:
            yield value


@subquery
class SubGraph(object):

    def __init__(self, graph, node):
        self.graph = graph
        self.node = node

        types = graph_types(graph)
        types['this'] = types[node]  # make an alias
        self.types = types

    def __call__(self, queue, ctx, task_set, node, fields, ids):
        graph_fields = [node.fields_map[f.name] for f in fields]
        fn_env = {f.__def_name__: f.__def_body__
                  for f in chain.from_iterable(e.functions
                                               for e in graph_fields)}

        this_link = Link(THIS_LINK_NAME, Sequence[TypeRef[self.node]],
                         None, requires=None)

        reqs = merge(f.reqs for f in graph_fields)
        procs = [f.proc for f in graph_fields]
        options = [list(_yield_options(qf, gf))
                   for qf, gf in zip(fields, graph_fields)]

        this_req = reqs.fields_map['this'].node
        other_reqs = query.Node([r for r in reqs.fields
                                 if r.name != 'this'])

        q = Query(queue, task_set, self.graph, ctx)
        q.process_link(self.graph.root, this_link, this_req, None, ids)
        q.process_node(self.graph.root, other_reqs, None)
        return _create_result_proc(q, fn_env, node, fields, procs,
                                   options, ids)



from collections import Sequence

from ..query import QueryVisitor
from ..compat import string_types
from ..protobuf import query_pb2


class Exporter(QueryVisitor):

    def __init__(self, node):
        self.stack = [node]

    def _populate_options(self, pb_obj, options):
        for k, v in options.items():
            if isinstance(v, int):
                pb_obj.options[k].integer = v
            elif isinstance(v, string_types):
                pb_obj.options[k].string = v
            elif isinstance(v, Sequence):
                if all(isinstance(i, int) for i in v):
                    pb_obj.options[k].repeated_integer.items[:] = v
                elif all(isinstance(i, string_types) for i in v):
                    pb_obj.options[k].repeated_string.items[:] = v
                else:
                    raise TypeError('Invalid option items type: {}={!r}'
                                    .format(k, v))
            else:
                raise TypeError('Invalid option value type: {}={!r}'
                                .format(k, v))

    def visit_field(self, obj):
        field = self.stack[-1].items.add().field
        field.name = obj.name
        if obj.options is not None:
            self._populate_options(field, obj.options)

    def visit_link(self, obj):
        link = self.stack[-1].items.add().link
        link.name = obj.name
        if obj.options is not None:
            self._populate_options(link, obj.options)
        self.stack.append(link.node)
        try:
            self.visit(obj.node)
        finally:
            self.stack.pop()

    def visit_node(self, obj):
        for item in obj.fields:
            self.visit(item)


def export(query):
    node = query_pb2.Node()
    Exporter(node).visit(query)
    return node



from ..edn import loads, Dict, List, Keyword, Tuple
from ..query import Node, Link, Field
from ..compat import text_type


def _get_options(value):
    if len(value) < 2:
        raise TypeError('Missing options argument')
    elif len(value) > 2:
        raise TypeError('More arguments than expected')

    keyword_value, options_value = value
    if not isinstance(keyword_value, Keyword):
        raise TypeError('Names should be specified as keywords, not as {!r}'
                        .format(type(keyword_value)))

    if not isinstance(options_value, Dict):
        raise TypeError('Options should be specified as mapping, not as {!r}'
                        .format(type(options_value)))

    non_keyword = set((k, type(k)) for k in options_value.keys()
                      if not isinstance(k, Keyword))
    if non_keyword:
        keys_repr = ' '.join('{} {!r}'.format(k, t) for k, t in non_keyword)
        raise TypeError('Option names should be specified as keywords: {}'
                        .format(keys_repr))

    name = text_type(keyword_value)
    options = {text_type(k): v for k, v in options_value.items()}
    return name, options


def _extract(values):
    for value in values:
        if isinstance(value, Tuple):
            name, options = _get_options(value)
            yield Field(name, options)
        elif isinstance(value, Keyword):
            yield Field(text_type(value))
        elif isinstance(value, Dict):
            for key, val in value.items():
                if isinstance(key, Tuple):
                    name, options = _get_options(key)
                elif isinstance(key, Keyword):
                    name = text_type(key)
                    options = None
                else:
                    raise TypeError('Link name defined not as keyword, '
                                    'but as {!r}'.format(key))
                yield Link(name, transform(val), options)
        else:
            raise TypeError('Invalid node member: {!r}'.format(value))


def transform(value):
    if isinstance(value, List):
        return Node(list(_extract(value)))
    else:
        raise TypeError('Node should be defined as vector, '
                        '{!r} provided instead'.format(value))


def read(src):
    edn_ast = loads(src)
    return transform(edn_ast)



import sys
import ast as _ast
import inspect


PY3 = sys.version_info[0] == 3
PY35 = sys.version_info >= (3, 5)
PY36 = sys.version_info >= (3, 6)


def with_metaclass(meta, *bases):
    """Create a base class with a metaclass."""
    # This requires a bit of explanation: the basic idea is to make a dummy
    # metaclass for one level of class instantiation that replaces itself with
    # the actual metaclass.
    class metaclass(meta):

        def __new__(cls, name, this_bases, d):
            return meta(name, bases, d)
    return type.__new__(metaclass, 'temporary_class', (), {})


class _AST(object):

    def __getattr__(self, name):
        return getattr(_ast, name)

    if PY3:
        @staticmethod
        def arguments(args, vararg, kwarg, defaults):
            return _ast.arguments(args, vararg, [], [], kwarg, defaults)

        @staticmethod
        def arg(arg):
            return _ast.arg(arg, None)

        if PY35:
            @staticmethod
            def Call(func, args, keywords, starargs, kwargs):
                return _ast.Call(func, args, keywords)
        else:
            Call = _ast.Call

    else:
        @staticmethod
        def Name(id, ctx):
            return _ast.Name(str(id), ctx)

        @staticmethod
        def Attribute(value, attr, ctx):
            return _ast.Attribute(value, str(attr), ctx)

        @staticmethod
        def arguments(args, vararg, kwarg, defaults):
            return _ast.arguments(args, vararg, kwarg, defaults)

        @staticmethod
        def arg(arg):
            return _ast.Name(str(arg), _ast.Param())

        Call = _ast.Call


ast = _AST()


if PY3:
    text_type = str
    string_types = str,

    def qualname(fn):
        if inspect.ismethod(fn):
            return fn.__func__.__qualname__
        else:
            return fn.__qualname__

else:
    text_type = unicode  # noqa
    string_types = basestring,  # noqa

    def qualname(fn):
        if inspect.ismethod(fn):
            return '{}.{}'.format(fn.im_class.__name__, fn.im_func.__name__)
        else:
            return fn.__name__



from __future__ import absolute_import

from hiku.sources import sqlalchemy as _sa


class FieldsQuery(_sa.FieldsQuery):

    async def __call__(self, ctx, fields_, ids):
        if not ids:
            return []

        expr, result_proc = self.__select_expr__(fields_, ids)

        sa_engine = ctx[self.sa_engine_ctx_var]
        async with sa_engine.acquire() as connection:
            res = await connection.execute(expr)
            rows = await res.fetchall()

        return result_proc(rows)


Field = _sa.Field


class LinkQuery(_sa.LinkQuery):

    async def __call__(self, ctx, ids):
        expr, result_proc = self.__select_expr__(ids)
        if expr is None:
            pairs = []
        else:
            sa_engine = ctx[self.sa_engine_ctx_var]
            async with sa_engine.acquire() as connection:
                res = await connection.execute(expr)
                rows = await res.fetchall()
            pairs = [(r.from_column, r.to_column) for r in rows]
        return result_proc(pairs)


Link = _sa.Link



from contextlib import contextmanager
from collections import Counter

from ..graph import GraphVisitor
from ..graph import AbstractNode, AbstractField, AbstractLink, AbstractOption

from .errors import Errors


def _get_duplicates(names):
    counter = Counter(names)
    return [k for k, v in counter.items() if v > 1]


def _format_names(names):
    return ', '.join('"{}"'.format(name) for name in names)


def _format_types(objects):
    return ', '.join(map(repr, set(type(obj) for obj in objects)))


class _NameFormatter(GraphVisitor):

    def visit_node(self, obj):
        return obj.name

    def visit_root(self, obj):
        return 'root'

    def visit_link(self, obj):
        return '.{}'.format(obj.name)

    def visit_field(self, obj):
        return '.{}'.format(obj.name)

    def visit_option(self, obj):
        return ':{}'.format(obj.name)


class GraphValidator(GraphVisitor):
    _name_formatter = _NameFormatter()
    _graph_accept_types = (AbstractNode,)
    _node_accept_types = (AbstractNode, AbstractField, AbstractLink)
    _link_accept_types = (AbstractOption,)

    def __init__(self, graph):
        self.graph = graph
        self.errors = Errors()
        self._ctx = []

    @contextmanager
    def push_ctx(self, obj):
        self._ctx.append(obj)
        try:
            yield
        finally:
            self._ctx.pop()

    @property
    def ctx(self):
        return self._ctx[-1]

    def _format_path(self, obj=None):
        path = self._ctx[1:] + ([obj] if obj is not None else [])
        return ''.join(self._name_formatter.visit(i) for i in path)

    def visit_option(self, obj):
        # TODO: check option default value according to the option type
        pass

    def visit_link(self, obj):
        invalid = [f for f in obj.options
                   if not isinstance(f, self._link_accept_types)]
        if invalid:
            self.errors.report(
                'Invalid types provided as link "{}" options: {}'
                .format(self._format_path(obj), _format_types(invalid))
            )
            return

        with self.push_ctx(obj):
            super(GraphValidator, self).visit_link(obj)

        if obj.node not in self.graph.nodes_map:
            self.errors.report(
                'Link "{}" points to the missing node "{}"'
                .format(self._format_path(obj), obj.node)
            )

        if obj.requires is not None:
            if obj.requires not in self.ctx.fields_map:
                self.errors.report(
                    'Link "{}" requires missing field "{}" in the "{}" node'
                    .format(obj.name, obj.requires, self._format_path())
                )

    def _generic_visit_node(self, obj, node_name):
        invalid = [f for f in obj.fields
                   if not isinstance(f, self._node_accept_types)]
        if invalid:
            self.errors.report(
                'Node can not contain these types: {} in node "{}"'
                .format(_format_types(invalid), node_name)
            )
            return

        with self.push_ctx(obj):
            for item in obj.fields:
                self.visit(item)

        duplicates = _get_duplicates(e.name for e in obj.fields)
        if duplicates:
            self.errors.report('Duplicated names found in the "{}" '
                               'node: {}'
                               .format(node_name, _format_names(duplicates)))

    def visit_node(self, obj):
        self._generic_visit_node(obj, obj.name)
        nodes = [f.name for f in obj.fields if isinstance(f, AbstractNode)]
        if nodes:
            self.errors.report(
                'Node can not be defined in the non-root node: '
                '{} in "{}"'
                .format(_format_names(nodes), obj.name)
            )

    def visit_root(self, obj):
        self._generic_visit_node(obj, 'root')

    def visit_graph(self, obj):
        invalid = [f for f in obj.items
                   if not isinstance(f, self._graph_accept_types)]
        if invalid:
            self.errors.report(
                'Graph can not contain these types: {}'
                .format(_format_types(invalid))
            )
            return

        with self.push_ctx(obj):
            self.visit(obj.root)
            for item in obj.nodes:
                self.visit(item)

        duplicates = _get_duplicates(e.name for e in obj.nodes)
        if duplicates:
            self.errors.report('Duplicated node names found in the graph: {}'
                               .format(_format_names(duplicates)))



import inspect

from functools import partial
from itertools import chain
from collections import defaultdict

from . import query
from .graph import Link, Node, Maybe, One, Many, Nothing, Field
from .result import Result
from .executors.queue import Workflow, Queue


class SplitPattern(query.QueryVisitor):

    def __init__(self, node):
        self._node = node
        self._fields = []
        self._links = []
        self._nodes = []

    def split(self, pattern):
        for item in pattern.fields:
            self.visit(item)
        return self._fields, self._links, self._nodes

    def visit_node(self, obj):
        raise ValueError('Unexpected value: {!r}'.format(obj))

    def visit_field(self, obj):
        self._fields.append((self._node.fields_map[obj.name], obj))

    def visit_link(self, obj):
        graph_obj = self._node.fields_map[obj.name]
        if isinstance(graph_obj, Link):
            if graph_obj.requires:
                self._fields.append((self._node.fields_map[graph_obj.requires],
                                     query.Field(graph_obj.requires)))
            self._links.append((graph_obj, obj))
        elif isinstance(graph_obj, Node):
            self._nodes.append(obj)
        else:
            assert isinstance(graph_obj, Field), type(graph_obj)
            # `obj` here is a link, but this link is treated as a complex field
            self._fields.append((graph_obj, obj))


def store_fields(result, node, fields, ids, query_result):
    names = [f.name for f in fields]
    if node.name is not None:
        if ids is not None:
            node_index = result.index.setdefault(node.name, {})
            for i, row in zip(ids, query_result):
                node_data = node_index.setdefault(i, {})
                node_data.update(zip(names, row))
        else:
            node_data = result.root.setdefault(node.name, {})
            node_data.update(zip(names, query_result))
    else:
        result.root.update(zip(names, query_result))


def link_reqs(result, node, link, ids):
    if node.name is not None:
        if ids is not None:
            return [result.index[node.name][i][link.requires] for i in ids]
        else:
            return result.root[node.name][link.requires]
    else:
        return result.root[link.requires]


def link_ref_maybe(result, link, ident):
    return None if ident is Nothing else result.ref(link.node, ident)


def link_ref_one(result, link, ident):
    assert ident is not Nothing
    return result.ref(link.node, ident)


def link_ref_many(result, link, idents):
    return [result.ref(link.node, i) for i in idents]


_LINK_REF_MAKER = {
    Maybe: link_ref_maybe,
    One: link_ref_one,
    Many: link_ref_many,
}


def store_links(result, node, link, ids, query_result):
    field_val = partial(_LINK_REF_MAKER[link.type_enum], result, link)
    if node.name is not None:
        if ids is not None:
            node_index = result.index.setdefault(node.name, {})
            for i, res in zip(ids, query_result):
                node_data = node_index.setdefault(i, {})
                node_data[link.name] = field_val(res)
        else:
            node_data = result.root.setdefault(node.name, {})
            node_data[link.name] = field_val(query_result)
    else:
        result.root[link.name] = field_val(query_result)


def link_result_to_ids(is_list, link_type, result):
    if is_list:
        if link_type is Maybe:
            return [i for i in result if i is not Nothing]
        elif link_type is One:
            if any(i is Nothing for i in result):
                raise TypeError('Non-optional link should not return Nothing: '
                                '{!r}'.format(result))
            return result
        elif link_type is Many:
            return list(chain.from_iterable(result))
    else:
        if link_type is Maybe:
            return [] if result is Nothing else [result]
        elif link_type is One:
            if result is Nothing:
                raise TypeError('Non-optional link should not return Nothing')
            return [result]
        elif link_type is Many:
            return result
    raise TypeError(repr([is_list, link_type]))


def get_options(graph_obj, query_obj):
    _options = query_obj.options or {}
    options = {}
    for opt in graph_obj.options:
        opt_value = _options.get(opt.name, opt.default)
        if opt_value is Nothing:
            raise TypeError('Required option "{}" for {!r} was not provided'
                            .format(opt.name, graph_obj))
        options[opt.name] = opt_value
    return options


class Query(Workflow):

    def __init__(self, queue, task_set, graph, ctx):
        self._queue = queue
        self._task_set = task_set
        self.graph = graph
        self._ctx = ctx
        self._result = Result()

    def _submit(self, func, *args, **kwargs):
        if _do_pass_context(func):
            return self._task_set.submit(func, self._ctx, *args, **kwargs)
        else:
            return self._task_set.submit(func, *args, **kwargs)

    def result(self):
        return self._result

    def process_node(self, node, pattern, ids):
        fields, links, nodes = SplitPattern(node).split(pattern)

        assert not (node.name and nodes), 'Nested nodes are not supported yet'
        for link in nodes:
            self.process_node(node.fields_map[link.name], link.node, None)

        to_func = {}
        from_func = defaultdict(list)
        for graph_field, query_field in fields:
            to_func[graph_field.name] = graph_field.func
            from_func[graph_field.func].append(query_field)

        # schedule fields resolve
        to_fut = {}
        for func, func_fields in from_func.items():
            if _is_subquery(func):
                task_set = self._queue.fork(self._task_set)
                if ids is not None:
                    result_proc = func(self._queue, self._ctx, task_set,
                                       node, func_fields, ids)
                else:
                    result_proc = func(self._queue, self._ctx, task_set,
                                       node, func_fields)
                to_fut[func] = task_set
                self._queue.add_callback(task_set, (
                    lambda:
                    result_proc(self._result)
                ))
            else:
                if ids is not None:
                    fut = self._submit(func, func_fields, ids)
                else:
                    fut = self._submit(func, func_fields)
                to_fut[func] = fut
                self._queue.add_callback(fut, (
                    lambda _fut=fut, _func_fields=func_fields:
                    store_fields(self._result, node, _func_fields, ids,
                                 _fut.result())
                ))

        # schedule link resolve
        for graph_link, query_link in links:
            if graph_link.requires:
                fut = to_fut[to_func[graph_link.requires]]
                self._queue.add_callback(fut, (
                    lambda _gl=graph_link, _ql=query_link:
                    self._process_node_link(node, _gl, _ql, ids)
                ))
            else:
                if graph_link.options:
                    options = get_options(graph_link, query_link)
                    fut = self._submit(graph_link.func, options)
                else:
                    fut = self._submit(graph_link.func)
                self._queue.add_callback(fut, (
                    lambda _fut=fut, _gl=graph_link, _qe=query_link.node:
                    self.process_link(node, _gl, _qe, ids, _fut.result())
                ))

    def _process_node_link(self, node, graph_link, query_link, ids):
        reqs = link_reqs(self._result, node, graph_link, ids)
        if graph_link.options:
            options = get_options(graph_link, query_link)
            fut = self._submit(graph_link.func, reqs, options)
        else:
            fut = self._submit(graph_link.func, reqs)
        self._queue.add_callback(fut, (
            lambda:
            self.process_link(node, graph_link, query_link.node, ids,
                              fut.result())
        ))

    def process_link(self, node, graph_link, query_node, ids, result):
        if inspect.isgenerator(result):
            result = list(result)
        store_links(self._result, node, graph_link, ids, result)
        to_ids = link_result_to_ids(ids is not None, graph_link.type_enum,
                                    result)
        if to_ids:
            self.process_node(self.graph.nodes_map[graph_link.node], query_node,
                              to_ids)


def pass_context(func):
    func.__pass_context__ = True
    return func


def _do_pass_context(func):
    return getattr(func, '__pass_context__', False)


def subquery(func):
    func.__subquery__ = True
    return func


def _is_subquery(func):
    return getattr(func, '__subquery__', False)


class Context(object):

    def __init__(self, mapping):
        self.__mapping = mapping

    def __getitem__(self, item):
        try:
            return self.__mapping[item]
        except KeyError:
            raise KeyError('Context variable {!r} is not specified '
                           'in the query context'.format(item))


class Engine(object):

    def __init__(self, executor):
        self.executor = executor

    def execute(self, graph, pattern, ctx=None):
        queue = Queue(self.executor)
        task_set = queue.fork(None)
        q = Query(queue, task_set, graph, Context(ctx or {}))
        q.process_node(q.graph.root, pattern, None)
        return self.executor.process(queue, q)



from __future__ import absolute_import

from contextlib import contextmanager

from ..types import Record, RecordMeta, OptionalMeta, SequenceMeta, MappingMeta
from ..types import GenericMeta, Unknown
from ..expr.checker import GraphTypes

from .types import TypeDef


# TODO: revisit this
_CONTAINER_TYPES = (
    OptionalMeta,
    SequenceMeta,
    MappingMeta,
    RecordMeta,
)


class TypeDoc(object):

    def __init__(self, type_, description):
        self.__type__ = type_
        self.__type_description__ = description

    def __getattr__(self, name):
        return getattr(self.__type, name)


class GraphTypesEx(GraphTypes):

    def visit(self, obj):
        t = super(GraphTypesEx, self).visit(obj)
        if isinstance(t, GenericMeta) and obj.description is not None:
            t = TypeDoc(t, obj.description)
        return t

    def visit_graph(self, obj):
        types_map = super(GraphTypesEx, self).visit_graph(obj)
        return [TypeDef[n, t] for n, t in types_map.items()
                if t is not Unknown]

    def visit_node(self, obj):
        record = super(GraphTypesEx, self).visit_node(obj)
        return Record[[(n, t) for n, t in record.__field_types__.items()
                       if t is not Unknown]]

    def visit_root(self, obj):
        record = super(GraphTypesEx, self).visit_root(obj)
        return Record[[(n, t) for n, t in record.__field_types__.items()
                       if t is not Unknown]]


class _LinePrinter(object):

    def visit(self, type_):
        return type_.accept(self)

    def visit_boolean(self, type_):
        return 'Boolean'

    def visit_string(self, type_):
        return 'String'

    def visit_integer(self, type_):
        return 'Integer'

    def visit_typeref(self, type_):
        return type_.__type_name__

    def visit_unknown(self, type_):
        return 'Unknown'


class _IndentedPrinter(object):
    _indent_size = 2

    def __init__(self):
        self._indent = 0
        self._buffer = []
        self._descriptions = {}

    @contextmanager
    def _add_indent(self):
        self._indent += 1
        yield
        self._indent -= 1

    def _newline(self):
        self._buffer.append('')

    def _print_call(self, line):
        self._buffer.append((' ' * self._indent_size * self._indent) + line)

    def _print_arg(self, type_):
        if isinstance(type_, TypeDoc):
            self._descriptions[len(self._buffer) - 1] = \
                type_.__type_description__
            type_ = type_.__type__
        if isinstance(type_, _CONTAINER_TYPES):
            with self._add_indent():
                self.visit(type_)
        else:
            self._buffer[-1] += ' ' + _LinePrinter().visit(type_)

    def _iter_lines(self):
        for i, line in enumerate(self._buffer):
            if i in self._descriptions:
                yield line + '  ; {}'.format(self._descriptions[i])
            else:
                yield line

    @classmethod
    def dumps(cls, types):
        printer = cls()
        for i, type_ in enumerate(types):
            if i > 0:
                printer._newline()
            printer.visit(type_)
        return '\n'.join(printer._iter_lines()) + '\n'

    def visit(self, type_):
        type_.accept(self)

    def visit_typedef(self, type_):
        self._print_call('type {}'.format(type_.__type_name__))
        self._print_arg(type_.__type__)

    def visit_record(self, type_):
        self._print_call('Record')
        with self._add_indent():
            for name, field_type in type_.__field_types__.items():
                self._print_call(':{}'.format(name))
                self._print_arg(field_type)

    def visit_sequence(self, type_):
        self._print_call('List')
        self._print_arg(type_.__item_type__)

    def visit_mapping(self, type_):
        self._print_call('Dict')
        self._print_arg(type_.__key_type__)
        self._print_arg(type_.__value_type__)

    def visit_optional(self, type_):
        self._print_call('Option')
        self._print_arg(type_.__type__)


def dumps(graph):
    types = GraphTypesEx().visit(graph)
    return _IndentedPrinter.dumps(types)



from concurrent.futures import wait, FIRST_COMPLETED


class ThreadsExecutor(object):

    def __init__(self, pool):
        self._pool = pool

    def submit(self, fn, *args, **kwargs):
        return self._pool.submit(fn, *args, **kwargs)

    def process(self, queue, workflow):
        while queue.__futures__:
            done, _ = wait(queue.__futures__, return_when=FIRST_COMPLETED)
            queue.progress(done)
        return workflow.result()



from __future__ import absolute_import

from itertools import chain

from graphql.language import ast
from graphql.language.parser import parse

from ..query import Node, Field, Link


class NodeVisitor(object):

    def visit(self, obj):
        if not isinstance(obj, ast.Node):
            raise TypeError('Unknown node type: {!r}'.format(obj))
        visit_method = getattr(self, 'visit_{}'.format(obj.__class__.__name__),
                               None)
        if visit_method is None:
            raise NotImplementedError('Not implemented node type: {!r}'
                                      .format(obj))
        return visit_method(obj)

    def visit_Document(self, obj):
        for definition in obj.definitions:
            self.visit(definition)

    def visit_OperationDefinition(self, obj):
        self.visit(obj.selection_set)

    def visit_FragmentDefinition(self, obj):
        self.visit(obj.selection_set)

    def visit_SelectionSet(self, obj):
        for i in obj.selections:
            self.visit(i)

    def visit_Field(self, obj):
        pass

    def visit_FragmentSpread(self, obj):
        pass

    def visit_InlineFragment(self, obj):
        self.visit(obj.selection_set)


class FragmentsCollector(NodeVisitor):

    def __init__(self):
        self.fragments_map = {}

    def visit_OperationDefinition(self, obj):
        pass  # not interested in operations here

    def visit_FragmentDefinition(self, obj):
        if obj.name.value in self.fragments_map:
            raise TypeError('Duplicated fragment name: "{}"'
                            .format(obj.name.value))
        self.fragments_map[obj.name.value] = obj


class SelectionSetVisitMixin(object):

    def transform_fragment(self, name):
        raise NotImplementedError

    @property
    def query_variables(self):
        raise NotImplementedError

    @property
    def query_name(self):
        raise NotImplementedError

    def lookup_variable(self, name):
        try:
            return self.query_variables[name]
        except KeyError:
            raise TypeError('Variable ${} is not defined in query {}'
                            .format(name, self.query_name))

    def visit_SelectionSet(self, obj):
        for i in obj.selections:
            for j in self.visit(i):
                yield j

    def visit_Field(self, obj):
        if obj.alias is not None:
            raise TypeError('Field aliases are not supported: {!r}'
                            .format(obj))
        options = {arg.name.value: self.visit(arg.value)
                   for arg in obj.arguments}
        if obj.selection_set is None:
            yield Field(obj.name.value, options or None)
        else:
            node = Node(list(self.visit(obj.selection_set)))
            yield Link(obj.name.value, node, options or None)

    def visit_Variable(self, obj):
        return self.lookup_variable(obj.name.value)

    def _visit_scalar(self, obj):
        return obj.value

    def visit_IntValue(self, obj):
        return int(obj.value)

    def visit_FloatValue(self, obj):
        return float(obj.value)

    def visit_StringValue(self, obj):
        return obj.value

    def visit_BooleanValue(self, obj):
        return obj.value

    def visit_EnumValue(self, obj):
        return obj.value

    def visit_ListValue(self, obj):
        return [self.visit(i) for i in obj.values]

    def visit_ObjectValue(self, obj):
        return {f.name.value: self.visit(f.value) for f in obj.fields}

    def visit_FragmentSpread(self, obj):
        assert not obj.directives, obj.directives
        for i in self.transform_fragment(obj.name.value):
            yield i

    def visit_InlineFragment(self, obj):
        for i in self.visit(obj.selection_set):
            yield i


class FragmentsTransformer(SelectionSetVisitMixin, NodeVisitor):
    query_name = None
    query_variables = None

    def __init__(self, document, query_name, query_variables):
        collector = FragmentsCollector()
        collector.visit(document)
        self.query_name = query_name
        self.query_variables = query_variables
        self.fragments_map = collector.fragments_map
        self.cache = {}
        self.pending_fragments = set()

    def transform_fragment(self, name):
        return self.visit(self.fragments_map[name])

    def visit_OperationDefinition(self, obj):
        pass  # not interested in operations here

    def visit_FragmentDefinition(self, obj):
        if obj.name.value in self.cache:
            return self.cache[obj.name.value]
        else:
            if obj.name.value in self.pending_fragments:
                raise TypeError('Cyclic fragment usage: "{}"'
                                .format(obj.name.value))
            self.pending_fragments.add(obj.name.value)
            try:
                selection_set = list(self.visit(obj.selection_set))
            finally:
                self.pending_fragments.discard(obj.name.value)
            self.cache[obj.name.value] = selection_set
            return selection_set


class GraphQLTransformer(SelectionSetVisitMixin, NodeVisitor):
    query_name = None
    query_variables = None
    fragments_transformer = None

    def __init__(self, document, variables=None):
        self.document = document
        self.variables = variables

    @classmethod
    def transform(cls, document, variables=None):
        visitor = cls(document, variables)
        return visitor.visit(document)

    def transform_fragment(self, name):
        return self.fragments_transformer.transform_fragment(name)

    def visit_Document(self, obj):
        queries = list(chain.from_iterable(self.visit(i)
                                           for i in obj.definitions))
        if len(queries) != 1:
            raise TypeError('Only single operation per document is '
                            'supported, {} operations was provided'
                            .format(len(obj.definitions)))
        return queries[0]

    def visit_OperationDefinition(self, obj):
        if obj.operation != 'query':
            raise TypeError('Only "query" operations are supported, '
                            '"{}" operation was provided'
                            .format(obj.operation))
        assert obj.operation == 'query', obj.operation

        variables = self.variables or {}
        query_name = obj.name.value if obj.name else '<unnamed>'
        query_variables = {}
        for var_defn in obj.variable_definitions or ():
            name = var_defn.variable.name.value
            try:
                value = variables[name]
            except KeyError:
                if var_defn.default_value is not None:
                    value = self.visit(var_defn.default_value)
                else:
                    raise TypeError('Variable "{}" is not provided for query {}'
                                    .format(name, query_name))
            query_variables[name] = value

        self.query_name = query_name
        self.query_variables = query_variables
        self.fragments_transformer = FragmentsTransformer(self.document,
                                                          self.query_name,
                                                          self.query_variables)
        try:
            node = Node(list(self.visit(obj.selection_set)))
        finally:
            self.query_name = None
            self.query_variables = None
            self.fragments_transformer = None
        yield node

    def visit_FragmentDefinition(self, obj):
        return []  # not interested in fragments here


def read(src, variables=None):
    doc = parse(src)
    return GraphQLTransformer.transform(doc, variables)



"""
    hiku.graph
    ~~~~~~~~~~

    Graphs are defined by nodes, fields and links. Simple functions
    are used to fetch any data from any data source.

"""
from abc import ABCMeta, abstractmethod
from itertools import chain
from collections import OrderedDict

from .types import OptionalMeta, SequenceMeta, TypeRefMeta
from .utils import kw_only, cached_property, const
from .compat import with_metaclass

Maybe = const('Maybe')

One = const('One')

Many = const('Many')

#: Special constant that is used by links with :py:class:`~hiku.types.Optional`
#: type in order to indicate that there is nothing to link to
Nothing = const('Nothing')


class AbstractNode(with_metaclass(ABCMeta)):

    @abstractmethod
    def accept(self, visitor):
        pass


class AbstractOption(AbstractNode):
    pass


class Option(AbstractOption):
    """Defines an option, used to customize results of the fields and links

    Options without default value are **required**.

    Example of a required option::

        Option('id', Integer)

    Example of an optional option::

        Option('size', Integer, default=100)

    """
    def __init__(self, name, type_, **kwargs):
        """
        :param name: name of the option
        :param type_: type of the option or ``None``
        :param kw-only,optional default: default option value
        """
        self.name = name
        self.type = type_
        self.default, = kw_only(self.__init__, kwargs, [],
                                [('default', Nothing)])

    def __repr__(self):
        return '{}({!r}, {!r}, ...)'.format(self.__class__.__name__,
                                            self.name, self.type)

    def accept(self, visitor):
        return visitor.visit_option(self)


class AbstractField(AbstractNode):
    pass


class Field(AbstractField):
    """Defines a field of the node

    Example::

        graph = Graph([
            Node('user', [
                Field('name', String, func),
            ]),
        ])

    Example with options::

        graph = Graph([
            Root([
                Field('lorem-ipsum', String, func,
                      options=[Option('words', Integer, default=50)]),
            ]),
        ])

    Data loading protocol::

        # root node fields
        def func(fields) -> List[T]

        # non-root node fields
        def func(fields, ids) -> List[List[T]]

    Where:

    - ``fields`` - list of :py:class:`hiku.query.Field`
    - ``ids`` - list node identifiers

    """
    def __init__(self, name, type_, func, **kwargs):
        """
        :param str name: name of the field
        :param type_: type of the field or ``None``
        :param func: function to load field's data
        :param kw-only,optional options: list of acceptable options
        :param kw-only,optional description: description of the field
        """
        options, description = kw_only(self.__init__, kwargs, [],
                                       [('options', None),
                                        ('description', None)])

        self.name = name
        self.type = type_
        self.func = func
        self.options = options or ()
        self.description = description

    def __repr__(self):
        return '{}({!r}, {!r}, {!r})'.format(self.__class__.__name__, self.name,
                                             self.type, self.func)

    @cached_property
    def options_map(self):
        return OrderedDict((op.name, op) for op in self.options)

    def accept(self, visitor):
        return visitor.visit_field(self)


class AbstractLink(AbstractNode):
    pass


def get_type_enum(type_):
    if isinstance(type_, TypeRefMeta):
        return One, type_.__type_name__
    elif isinstance(type_, OptionalMeta):
        if isinstance(type_.__type__, TypeRefMeta):
            return Maybe, type_.__type__.__type_name__
    elif isinstance(type_, SequenceMeta):
        if isinstance(type_.__item_type__, TypeRefMeta):
            return Many, type_.__item_type__.__type_name__
    raise TypeError('Invalid type specified: {!r}'.format(type_))


class Link(AbstractLink):
    """Defines a link to the node

    Example::

        graph = Graph([
            Node('user', [...]),
            Root([
                Link('users', Sequence[TypeRef['user']], func, requires=None),
            ]),
        ])

    Example with requirements::

        graph = Graph([
            Node('character', [...]),
            Node('actor', [
                Field('id', Integer, field_func),
                Link('characters', Sequence[TypeRef['character']],
                     link_func, requires='id'),
            ])
        ])

    Requirements are needed when link points from non-root node.

    Example with options::

        graph = Graph([
            Node('user', [...]),
            Root([
                Link('users', Sequence[TypeRef['user']], func, requires=None,
                     options=[Option('limit', Integer, default=100)]),
            ]),
        ])

    Identifiers loading function protocol::

        # requires is None and type is TypeRef['foo']
        def func() -> T

        # requires is None and type is Optional[TypeRef['foo']]
        def func() -> Union[T, Nothing]

        # requires is None and type is Sequence[TypeRef['foo']]
        def func() -> List[T]

        # requires is not None and type is TypeRef['foo']
        def func(ids) -> List[T]

        # requires is not None and type is Optional[TypeRef['foo']]
        def func(ids) -> List[Union[T, Nothing]]

        # requires is not None and type is Sequence[TypeRef['foo']]
        def func(ids) -> List[List[T]]

    See also :py:const:`hiku.graph.Nothing`.

    If link was defined with options, then link function should accept one
    additional positional argument::

        # many to many link with options
        def func(ids, options) -> List[List[T]]

    Where ``options`` is a mapping ``str: value`` of provided in the query
    options.
    """
    def __init__(self, name, type_, func, **kwargs):
        """
        :param name: name of the link
        :param type_: type of the link
        :param func: function to load identifiers of the linked node
        :param kw-only requires: field name from the current node, required
                                      to compute identifiers of the linked node
        :param kw-only,optional options: list of acceptable options
        :param kw-only,optional description: description of the link
        """
        requires, options, description = \
            kw_only(self.__init__, kwargs, ['requires'],
                    [('options', None), ('description', None)])

        type_enum, node = get_type_enum(type_)

        self.name = name
        self.type = type_
        self.type_enum = type_enum
        self.node = node
        self.func = func
        self.requires = requires
        self.options = options or ()
        self.description = description

    def __repr__(self):
        return '{}({!r}, {!r}, {!r}, ...)'.format(self.__class__.__name__,
                                                  self.name, self.type,
                                                  self.func)

    @cached_property
    def options_map(self):
        return OrderedDict((op.name, op) for op in self.options)

    def accept(self, visitor):
        return visitor.visit_link(self)


class AbstractNode(AbstractNode):
    pass


class Node(AbstractNode):
    """Collection of the fields and links, which describes some entity and
    relations with other entities

    Example::

        graph = Graph([
            Node('user', [
                Field('id', Integer, field_func),
                Field('name', String, field_func),
                Link('roles', Sequence[TypeRef['role']],
                     link_func, requires='id'),
            ]),
        ])

    """
    def __init__(self, name, fields, **kwargs):
        """
        :param name: name of the node
        :param fields: list of fields and links
        :param kw-only,optional description: description of the node
        """
        self.name = name
        self.fields = fields
        self.description, = kw_only(self.__init__, kwargs, [],
                                    [('description', None)])

    def __repr__(self):
        return '{}({!r}, {!r}, ...)'.format(self.__class__.__name__, self.name,
                                            self.fields)

    @cached_property
    def fields_map(self):
        return OrderedDict((f.name, f) for f in self.fields)

    def accept(self, visitor):
        return visitor.visit_node(self)


class Root(Node):
    """Special implicit root node, starting point of the query execution

    Example::

        graph = Graph([
            Node('baz', [...]),
            Root([
                Field('foo', String, root_fields_func),
                Link('bar', Sequence[TypeRef['baz']],
                     to_baz_func, requires=None),
                Node('quux', [...]),
            ]),
        ])

    """
    def __init__(self, items):
        """
        :param items: list of fields, links and singleton nodes
        """
        super(Root, self).__init__(None, items)

    def __repr__(self):
        return '{}({!r}, {!r})'.format(self.__class__.__name__, self.name,
                                       self.fields)

    def accept(self, visitor):
        return visitor.visit_root(self)


class AbstractGraph(AbstractNode):
    pass


class Graph(AbstractGraph):
    """Collection of nodes - definition of the graph

    Example::

        graph = Graph([
            Node('foo', [...]),
            Node('bar', [...]),
            Root([...]),
        ])

    """
    def __init__(self, items):
        """
        :param items: list of nodes
        """
        self.items = items

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.items)

    @cached_property
    def root(self):
        return Root(list(chain.from_iterable(e.fields for e in self.items
                                             if e.name is None)))

    @cached_property
    def nodes(self):
        return [e for e in self.items if e.name is not None]

    @cached_property
    def nodes_map(self):
        return OrderedDict((e.name, e) for e in self.nodes)

    def accept(self, visitor):
        return visitor.visit_graph(self)


class AbstractGraphVisitor(with_metaclass(ABCMeta, object)):

    @abstractmethod
    def visit(self, obj):
        pass

    @abstractmethod
    def visit_option(self, obj):
        pass

    @abstractmethod
    def visit_field(self, obj):
        pass

    @abstractmethod
    def visit_link(self, obj):
        pass

    @abstractmethod
    def visit_node(self, obj):
        pass

    @abstractmethod
    def visit_root(self, obj):
        pass

    @abstractmethod
    def visit_graph(self, obj):
        pass


class GraphVisitor(AbstractGraphVisitor):

    def visit(self, obj):
        return obj.accept(self)

    def visit_option(self, obj):
        pass

    def visit_field(self, obj):
        for option in obj.options:
            self.visit(option)

    def visit_link(self, obj):
        for option in obj.options:
            self.visit(option)

    def visit_node(self, obj):
        for item in obj.fields:
            self.visit(item)

    def visit_root(self, obj):
        for item in obj.fields:
            self.visit(item)

    def visit_graph(self, obj):
        for item in obj.items:
            self.visit(item)


class GraphTransformer(AbstractGraphVisitor):

    def visit(self, obj):
        return obj.accept(self)

    def visit_option(self, obj):
        return Option(obj.name, obj.type, default=obj.default)

    def visit_field(self, obj):
        return Field(obj.name, obj.type, obj.func,
                     options=[self.visit(op) for op in obj.options],
                     description=obj.description)

    def visit_link(self, obj):
        return Link(obj.name, obj.type, obj.func,
                    requires=obj.requires,
                    options=[self.visit(op) for op in obj.options],
                    description=obj.description)

    def visit_node(self, obj):
        return Node(obj.name, [self.visit(f) for f in obj.fields],
                    description=obj.description)

    def visit_root(self, obj):
        return Root([self.visit(f) for f in obj.fields])

    def visit_graph(self, obj):
        return Graph([self.visit(node) for node in obj.items])



class FutureLike(object):

    def __init__(self, result):
        self._result = result

    def result(self):
        return self._result


class SyncExecutor(object):

    def submit(self, fn, *args, **kwargs):
        return FutureLike(fn(*args, **kwargs))

    def process(self, queue, workflow):
        while queue.__futures__:
            queue.progress(queue.__futures__)
        return workflow.result()



from __future__ import absolute_import

from collections import defaultdict

import sqlalchemy

from ..utils import kw_only
from ..types import String, Integer
from ..graph import Field as FieldBase, Link as LinkBase, get_type_enum
from ..graph import Nothing, Maybe, One, Many
from ..engine import pass_context


@pass_context
class FieldsQuery(object):

    def __init__(self, sa_engine_ctx_var, from_clause, primary_key=None):
        self.sa_engine_ctx_var = sa_engine_ctx_var
        self.from_clause = from_clause
        if primary_key is not None:
            self.primary_key = primary_key
        else:
            # currently only one column supported
            self.primary_key, = from_clause.primary_key

    def __select_expr__(self, fields_, ids):
        columns = [self.from_clause.c[f.name] for f in fields_]
        expr = (
            sqlalchemy.select([self.primary_key] + columns)
            .select_from(self.from_clause)
            .where(self.primary_key.in_(ids))
        )

        def result_proc(rows):
            rows_map = {row[self.primary_key]: [row[c] for c in columns]
                        for row in rows}

            nulls = [None for _ in fields_]
            return [rows_map.get(id_, nulls) for id_ in ids]

        return expr, result_proc

    def __call__(self, ctx, fields_, ids):
        if not ids:
            return []

        expr, result_proc = self.__select_expr__(fields_, ids)

        sa_engine = ctx[self.sa_engine_ctx_var]
        with sa_engine.connect() as connection:
            rows = connection.execute(expr).fetchall()

        return result_proc(rows)


def _translate_type(column):
    if isinstance(column.type, sqlalchemy.Integer):
        return Integer
    elif isinstance(column.type, sqlalchemy.Unicode):
        return String
    else:
        return None


class Field(FieldBase):

    def __init__(self, name, query, **kwargs):
        try:
            column = query.from_clause.c[name]
        except KeyError:
            raise ValueError('FromClause {} does not have a column named {}'
                             .format(query.from_clause, name))
        type_ = _translate_type(column)
        super(Field, self).__init__(name, type_, query, **kwargs)


def _to_maybe_mapper(pairs, values):
    mapping = dict(pairs)
    return [mapping.get(value, Nothing) for value in values]


def _to_one_mapper(pairs, values):
    mapping = dict(pairs)
    return [mapping[value] for value in values]


def _to_many_mapper(pairs, values):
    mapping = defaultdict(list)
    for from_value, to_value in pairs:
        mapping[from_value].append(to_value)
    return [mapping[value] for value in values]


_MAPPERS = {
    Maybe: _to_maybe_mapper,
    One: _to_one_mapper,
    Many: _to_many_mapper,
}


@pass_context
class LinkQuery(object):

    def __init__(self, type_, sa_engine_ctx_var, **kwargs):
        type_enum, node = get_type_enum(type_)

        from_column, to_column = kw_only(self.__init__, kwargs,
                                         ['from_column', 'to_column'])
        if from_column.table is not to_column.table:
            raise ValueError('from_column and to_column should belong to '
                             'the one table')

        self.type = type_
        self.type_enum = type_enum
        self.node = node
        self.sa_engine_ctx_var = sa_engine_ctx_var
        self.from_column = from_column
        self.to_column = to_column

    def __select_expr__(self, ids):
        filtered_ids = frozenset(filter(None, ids))
        if filtered_ids:
            expr = (
                sqlalchemy.select([self.from_column.label('from_column'),
                                   self.to_column.label('to_column')])
                .where(self.from_column.in_(filtered_ids))
            )
        else:
            expr = None

        def result_proc(pairs):
            mapper = _MAPPERS[self.type_enum]
            return mapper(pairs, ids)

        return expr, result_proc

    def __call__(self, ctx, ids):
        expr, result_proc = self.__select_expr__(ids)
        if expr is None:
            pairs = []
        else:
            sa_engine = ctx[self.sa_engine_ctx_var]
            with sa_engine.connect() as connection:
                pairs = connection.execute(expr).fetchall()
        return result_proc(pairs)


class Link(LinkBase):

    def __init__(self, name, query, **kwargs):
        super(Link, self).__init__(name, query.type, query, **kwargs)



from asyncio import wait, FIRST_COMPLETED


class AsyncIOExecutor(object):

    def __init__(self, loop):
        self._loop = loop

    def submit(self, fn, *args, **kwargs):
        return self._loop.create_task(fn(*args, **kwargs))

    async def process(self, queue, workflow):
        while queue.__futures__:
            done, _ = await wait(queue.__futures__,
                                 loop=self._loop,
                                 return_when=FIRST_COMPLETED)
            queue.progress(done)
        return workflow.result()
